import numpy as np
import timeit, sys, subprocess, os, glob
from libraries.magma import Magma

with open("fig_8a_red.csv", 'w') as f:
   f.write("minsupp\tmaxpsize\tavg\n")
   psize = 4
   for supp in (90, 80, 70, 60, 50, 40):
      sys.stderr.write(f"\rTesting old code on support = {supp}")
      sys.stderr.flush()
      m = Magma("data/aarhus/aarhus_complete", supp, psize)
      runtime1 = timeit.timeit(stmt = "m.run()", number = 1, globals = globals())
      runtime2 = timeit.timeit(stmt = "m.apply_rules()", number = 1, globals = globals())
      f.write("%s\t%s\t%s\n" % (supp, psize, runtime1 + runtime2))
      f.flush()

sys.stderr.write("\n")

with open("fig_8b_red.csv", 'w') as f:
   f.write("conf\tavg\n")
   for conf in (1, 0.95, 0.9, 0.8):
      supp = 50
      psize = 4
      sys.stderr.write(f"\rTesting old code on confidence = {conf}")
      sys.stderr.flush()
      m = Magma("data/aarhus/aarhus_complete", supp, psize, min_conf = conf)
      runtime1 = timeit.timeit(stmt = "m.run()", number = 1, globals = globals())
      runtime2 = timeit.timeit(stmt = "m.apply_rules()", number = 1, globals = globals())
      f.write("%s\t%s\n" % (conf, runtime1 + runtime2))
      f.flush()

sys.stderr.write("\n")

with open("fig_8c_red.csv", 'w') as f:
   f.write("minsupp\tmaxpsize\tavg\n")
   for psize in (3, 4, 5):
      supp = 90
      sys.stderr.write(f"\rTesting old code on pattern size = {psize}")
      sys.stderr.flush()
      m = Magma("data/aarhus/aarhus_complete", supp, psize)
      runtime1 = timeit.timeit(stmt = "m.run()", number = 1, globals = globals())
      runtime2 = timeit.timeit(stmt = "m.apply_rules()", number = 1, globals = globals())
      f.write("%s\t%s\t%s\n" % (supp, psize, runtime1 + runtime2))
      f.flush()

sys.stderr.write("\n")
stmt = "process = subprocess.Popen(bash_command.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE);output, error = process.communicate()"

with open("fig_8a_green.csv", 'w') as f:
   f.write("minsupp\tmaxpsize\tavg\n")
   for supp in (90, 80, 70, 60, 50, 40):
      psize = 4
      sys.stderr.write(f"\rTesting new code on support = {supp}")
      sys.stderr.flush()
      bash_command = "java -Xmx6g -cp libraries/moss_new.jar moss.Miner -inel -onel -x -m2 -n%s -s%s -k3 -Zmoss.rul -w0 -T data/aarhus/aarhus_complete.gspan moss.sub" % (psize, supp)
      runtime1 = timeit.timeit(stmt = stmt, number = 1, globals = globals())
      bash_command = "java -Xmx6g -cp libraries/moss_new.jar moss.Predictor -fnel moss.sub moss.rul data/aarhus/aarhus_complete.gspan moss.ped"
      runtime2 = timeit.timeit(stmt = stmt, number = 1, globals = globals())
      f.write("%s\t%s\t%s\n" % (supp, psize, runtime1 + runtime2))
      f.flush()

sys.stderr.write("\n")

with open("fig_8b_green.csv", 'w') as f:
   f.write("conf\tavg\n")
   for conf in (80, 90, 95, 100):
      supp = 50
      psize = 4
      sys.stderr.write(f"\rTesting new code on confidence = {conf}")
      sys.stderr.flush()
      bash_command = "java -Xmx6g -cp libraries/moss_new.jar moss.Miner -inel -onel -x -m2 -n%s -s%s -k3 -Zmoss.rul -w%s -T data/aarhus/aarhus_complete.gspan moss.sub" % (psize, supp, conf)
      runtime1 = timeit.timeit(stmt = stmt, number = 1, globals = globals())
      bash_command = "java -Xmx6g -cp libraries/moss_new.jar moss.Predictor -fnel moss.sub moss.rul data/aarhus/aarhus_complete.gspan moss.ped"
      runtime2 = timeit.timeit(stmt = stmt, number = 1, globals = globals())
      f.write("%s\t%s\n" % (conf, runtime1 + runtime2))
      f.flush()

sys.stderr.write("\n")

with open("fig_8c_green.csv", 'w') as f:
   f.write("minsupp\tmaxpsize\tavg\n")
   supp = 90
   for psize in (3, 4, 5):
      sys.stderr.write(f"\rTesting new code on pattern size = {psize}")
      sys.stderr.flush()
      bash_command = "java -Xmx6g -cp libraries/moss_new.jar moss.Miner -inel -onel -x -m2 -n%s -s%s -k3 -Zmoss.rul -w0 -T data/aarhus/aarhus_complete.gspan moss.sub" % (psize, supp)
      runtime1 = timeit.timeit(stmt = stmt, number = 1, globals = globals())
      bash_command = "java -Xmx6g -cp libraries/moss_new.jar moss.Predictor -fnel moss.sub moss.rul data/aarhus/aarhus_complete.gspan moss.ped"
      runtime2 = timeit.timeit(stmt = stmt, number = 1, globals = globals())
      f.write("%s\t%s\t%s\n" % (supp, psize, runtime1 + runtime2))
      f.flush()

sys.stderr.write("\nCleaning up...\n")

for f in glob.glob("data/aarhus/*.mossout"):
   os.remove(f)

for f in glob.glob("data/aarhus/*.rules"):
   os.remove(f)

for f in glob.glob("data/aarhus/*.csv"):
   os.remove(f)

os.remove("moss.sub")
os.remove("moss.rul")
os.remove("moss.ped")
