import subprocess, timeit, sys, os, glob
import numpy as np
from libraries.magma import Magma

with open("fig_9a.csv", 'w') as f:
   f.write("minsupp\tpreprocess\tmoss\tfind\tapply\n")
   psize = 4
   for supp in (90, 80, 70, 60, 50, 40):
      sys.stderr.write(f"\rTesting old code on support = {supp}")
      sys.stderr.flush()
      m = Magma("data/aarhus/aarhus_complete", supp, psize)
      preprocess = timeit.timeit(stmt = "m = Magma('data/aarhus/aarhus_complete', supp, psize)", number = 1, globals = globals())
      preprocess += timeit.timeit(stmt = "m.make_moss_input()", number = 1, globals = globals())
      run_moss = timeit.timeit(stmt = "m.run_moss()", number = 1, globals = globals())
      find_rules = timeit.timeit(stmt = "m.read_patterns()", number = 1, globals = globals())
      find_rules += timeit.timeit(stmt = "m.build_rules()", number = 1, globals = globals())
      apply_rules = timeit.timeit(stmt = "m.apply_rules()", number = 1, globals = globals())
      tot_runtime = preprocess + run_moss + find_rules + apply_rules
      f.write("%s\t%s\t%s\t%s\t%s\n" % (supp, preprocess / tot_runtime, run_moss / tot_runtime, find_rules / tot_runtime, apply_rules / tot_runtime))
      f.flush()

sys.stderr.write("\n")

with open("fig_9b.csv", 'w') as f:
   f.write("minsupp\tpreprocess\tmoss\tfind\tapply\n")
   for supp in (90, 80, 70, 60, 50, 40):
      psize = 4
      sys.stderr.write(f"\rTesting new code on support = {supp}")
      sys.stderr.flush()
      bash_command = "java -Xmx6g -cp libraries/moss_new.jar moss.Miner -inel -onel -x -m2 -n%s -s%s -k3 -Zmoss.rul -w0 -T data/aarhus/aarhus_complete.gspan moss.sub" % (psize, supp)
      process = subprocess.Popen(bash_command.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE)
      output, error = process.communicate()
      error = str(error)
      preprocess = 0
      run_moss = float(error.split('\\n')[2].split('[')[-1].split(']')[0].replace('s', ''))
      run_moss += float(error.split('\\n')[3].split('[')[-1].split(']')[0].replace('s', ''))
      run_moss += float(error.split('\\n')[6].split('[')[-1].split(']')[0].replace('s', ''))
      find_rules = float(error.split('\\n')[7].split('[')[-1].split(']')[0].replace('s', ''))
      bash_command = "java -Xmx6g -cp libraries/moss_new.jar moss.Predictor -fnel moss.sub moss.rul data/aarhus/aarhus_complete.gspan moss.ped"
      process = subprocess.Popen(bash_command.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE)
      output, error = process.communicate()
      error = str(error)
      apply_rules = float(error.split('\\n')[2].split('[')[-1].split(']')[0].replace('s', ''))
      apply_rules += float(error.split('\\n')[3].split('[')[-1].split(']')[0].replace('s', ''))
      apply_rules += float(error.split('\\n')[4].split('[')[-1].split(']')[0].replace('s', ''))
      apply_rules += float(error.split('\\n')[5].split('[')[-1].split(']')[0].replace('s', ''))
      apply_rules += float(error.split('\\n')[6].split('[')[-1].split(']')[0].replace('s', ''))
      tot_runtime = preprocess + run_moss + find_rules + apply_rules
      f.write("%s\t%s\t%s\t%s\t%s\n" % (supp, preprocess / tot_runtime, run_moss / tot_runtime, find_rules / tot_runtime, apply_rules / tot_runtime))
      f.flush()

sys.stderr.write("\nCleaning up...\n")

for f in glob.glob("data/aarhus/*.mossout"):
   os.remove(f)

for f in glob.glob("data/aarhus/*.rules"):
   os.remove(f)

for f in glob.glob("data/aarhus/*.csv"):
   os.remove(f)

os.remove("moss.sub")
os.remove("moss.rul")
os.remove("moss.ped")
