import random, sys, os, glob, subprocess, timeit
import numpy as np
import pandas as pd
import networkx as nx
from libraries.magma import Magma

def make_benchmark(nnodes, avgk, layers, labels):
   nodes = [random.randint(1, labels) for _ in range(nnodes)]
   edges_new = []
   edges_old = []
   for layer in range(layers):
      G = nx.powerlaw_cluster_graph(nnodes, int(round(avgk / 2)), 0.01)
      for e in G.edges:
         edges_new.append("e %s %s %s\n" % (e[0] + 1, e[1] + 1, layer))
         edges_old.append("%s\t%s\t%s\n" % (e[0] + 1, e[1] + 1, layer))
   with open("synth_%s_%s_%s_%s" % (nnodes, avgk, layers, labels), 'w') as f:
      for edge in edges_old:
         f.write(edge)
   with open("synth_%s_%s_%s_%s_nodeattrs" % (nnodes, avgk, layers, labels), 'w') as f:
      for node in range(nnodes):
         f.write("%s\t%s\n" % (node + 1, nodes[node]))
   with open("synth_%s_%s_%s_%s.gspan" % (nnodes, avgk, layers, labels), 'w') as f:
      for node in range(nnodes):
         f.write("v %s %s\n" % (node + 1, nodes[node]))
      for edge in edges_new:
         f.write(edge)

def run_new(nnodes, avgk, layers, labels):
   global bash_command
   stmt = "process = subprocess.Popen(bash_command.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE);output, error = process.communicate()"
   with open("new_runtimes.csv", 'a') as f:
      bash_command = f"java -Xmx6g -cp libraries/moss_new.jar moss.Miner -inel -onel -x -m2 -n3 -s17 -k3 -Zmoss.rul -w0 -T synth_{nnodes}_{avgk}_{layers}_{labels}.gspan moss.sub"
      runtime1 = timeit.timeit(stmt = stmt, number = 1, globals = globals())
      bash_command = f"java -Xmx6g -cp libraries/moss_new.jar moss.Predictor -fnel moss.sub moss.rul synth_{nnodes}_{avgk}_{layers}_{labels}.gspan moss.ped"
      runtime2 = timeit.timeit(stmt = stmt, number = 1, globals = globals())
      f.write(f"{nnodes}\t{avgk}\t{layers}\t{labels}\t{runtime1 + runtime2}\n")
      f.flush()

def run_old(nnodes, avgk, layers, labels):
   global m
   with open("old_runtimes.csv", 'a') as f:
      m = Magma(f"synth_{nnodes}_{avgk}_{layers}_{labels}", 17, 3)
      runtime1 = timeit.timeit(stmt = "m.run()", number = 1, globals = globals())
      runtime2 = timeit.timeit(stmt = "m.apply_rules()", number = 1, globals = globals())
      f.write(f"{nnodes}\t{avgk}\t{layers}\t{labels}\t{runtime1 + runtime2}\n")
      f.flush()

def run_benchmark(nnodes = 500, avgk = 8, layers = 7, labels = 4):
   sys.stderr.write(f"\r# nodes: {nnodes}\tAVG degree: {avgk}\t# layers: {layers}\t# labels: {labels}\trun: {run}")
   make_benchmark(nnodes, avgk, layers, labels)
   run_new(nnodes, avgk, layers, labels)
   run_old(nnodes, avgk, layers, labels)

m = None
bash_command = ""

for run in range(10):
   for nnodes in (100, 250, 500, 1000, 1500, 2000, 2500, 3000):
      run_benchmark(nnodes = nnodes)
   sys.stderr.write("\n")
   for avgk in (6, 8, 10, 12, 14):
      run_benchmark(avgk = avgk)
   sys.stderr.write("\n")
   for layers in (2, 3, 4, 5, 6, 7, 8, 9, 10):
      run_benchmark(layers = layers)
   sys.stderr.write("\n")
   for labels in (5, 4, 3, 2, 1):
      run_benchmark(labels = labels)
   sys.stderr.write("\n")

sys.stderr.write("Cleaning up...\n")

df_old = pd.read_csv("old_runtimes.csv", sep = "\t", header = None, names = ("nnodes", "avgk", "layers", "labels", "avg"))
df_new = pd.read_csv("new_runtimes.csv", sep = "\t", header = None, names = ("nnodes", "avgk", "layers", "labels", "avg"))

df_old.groupby(by = "nnodes")["avg"].mean().reset_index().to_csv("fig_10a_red.csv", sep = "\t", index = False)
df_new.groupby(by = "nnodes")["avg"].mean().reset_index().to_csv("fig_10a_green.csv", sep = "\t", index = False)

df_old.groupby(by = "layers")["avg"].mean().reset_index().to_csv("fig_10b_red.csv", sep = "\t", index = False)
df_new.groupby(by = "layers")["avg"].mean().reset_index().to_csv("fig_10b_green.csv", sep = "\t", index = False)

df_old.groupby(by = "avgk")["avg"].mean().reset_index().to_csv("fig_10c_red.csv", sep = "\t", index = False)
df_new.groupby(by = "avgk")["avg"].mean().reset_index().to_csv("fig_10c_green.csv", sep = "\t", index = False)

df_old.groupby(by = "labels")["avg"].mean().reset_index().to_csv("fig_10d_red.csv", sep = "\t", index = False)
df_new.groupby(by = "labels")["avg"].mean().reset_index().to_csv("fig_10d_green.csv", sep = "\t", index = False)

for f in glob.glob("synth*"):
   os.remove(f)

os.remove("moss.sub")
os.remove("moss.rul")
os.remove("moss.ped")

os.remove("old_runtimes.csv")
os.remove("new_runtimes.csv")
