import pandas as pd
import timeit, subprocess, os

def run_moss(network, supp):
   bash_command_bfs = f"java -Xmx6g -cp libraries/moss_new.jar moss.Miner -inel -onel -x -m2 -n4 -s{supp} -k3 -Zmoss.rul -w0 -T data/{network} moss_bfs.sub"
   stmt_bfs = "process = subprocess.Popen(bash_command_bfs.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE);output, error = process.communicate()"
   runtime_bfs = timeit.timeit(stmt = stmt_bfs, number = 1, globals = {"subprocess": subprocess, "bash_command_bfs": bash_command_bfs})
   bash_command_dfs = f"java -Xmx6g -cp libraries/moss_new.jar moss.Miner -inel -onel -g -x -m2 -n4 -s{supp} -k3 -Zmoss.rul -w0 -T data/{network} moss_dfs.sub"
   stmt_dfs = "process = subprocess.Popen(bash_command_dfs.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE);output, error = process.communicate()"
   runtime_dfs = timeit.timeit(stmt = stmt_dfs, number = 1, globals = {"subprocess": subprocess, "bash_command_dfs": bash_command_dfs})
   return runtime_bfs, runtime_dfs

runtimes = []
for supp in (90, 80, 70, 60, 50, 40):
   bfs, dfs = run_moss("aarhus/aarhus_complete.gspan", supp)
   runtimes.append((supp, bfs, dfs))

pd.DataFrame(data = runtimes, columns = ("supp", "bfs", "dfs")).to_csv("fig_11a.csv", sep = "\t", index = False)

runtimes = []
for supp in (50, 40, 30, 20, 10, 5):
   bfs, dfs = run_moss("phys/phys0.gspan", supp)
   runtimes.append((supp, bfs, dfs))

pd.DataFrame(data = runtimes, columns = ("supp", "bfs", "dfs")).to_csv("fig_11b.csv", sep = "\t", index = False)

os.remove("moss_bfs.sub")
os.remove("moss_dfs.sub")
os.remove("moss.rul")

